"""specific views for manipulating tables
"""


from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd

from loris import config
from loris.app.app import app
from loris.app.templates import form_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password
from loris.slack import execute_slack_message
from loris.io import string_dump, string_load


@app.route('/delete/<schema>/<table>',
           defaults={'subtable': None}, methods=['GET', 'POST'])
@app.route('/delete/<schema>/<table>/<subtable>', methods=['GET', 'POST'])
@login_required
def delete(schema, table, subtable):

    redirect_url = request.args.get(
        'target',
        url_for(
            'table', schema=schema, table=table, subtable=subtable
        )
    )
    # get id if it exists (will be restriction)
    print(request.args.get('_id', string_dump(None)))
    _id = string_load(request.args.get('_id', string_dump(None)))
    if _id is None:
        return redirect(redirect_url)

    # get table and create dynamic form
    table_class = getattr(config['schemata'][schema], table)
    # get table name
    table_name = '.'.join([schema, table])

    subtable = request.args.get('subtable', subtable)
    if not (subtable is None or subtable == 'None'):
        table_name = f'{table_name}.{subtable}'
        table_class = getattr(table_class, subtable)

    to_delete = table_class & _id
    # test if user is allowed to delete entry
    if not user_has_permission(to_delete, current_user.user_name):
        flash((
            f'User {current_user.user_name} is not'
            f' allowed to delete entry: {_id}'
        ), 'error')
        return redirect(redirect_url)

    conn = to_delete.connection
    if conn.in_transaction:
        flash((
            f"Entry `{_id}` is currently within a transaction! Cannot delete"
        ), 'error')
        return redirect(redirect_url)

    conn.start_transaction()
    try:
        count, message = to_delete._delete_cascade(return_message=True)
    except Exception as e:
        conn.cancel_transaction()
        flash((
            f"DataJointError: {e}"
        ), 'error')
        return redirect(redirect_url)

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        if submit == 'Delete' and count:
            if table_name in config.slack_tables:
                if 'delete' in config.slack_tables[table_name].get(
                    'upon', 'insert'
                ):
                    execute_slack_message(
                        to_delete.fetch1(),
                        config.slack_tables[table_name]
                    )
            conn.commit_transaction()
            # reset table (will not cascade)
            dynamicform, form = config.get_dynamicform(
                table_name, table_class, DynamicForm
            )
            dynamicform.reset()
            # redired to table
            flash(f'Entry deleted: {_id}', 'warning')
            return redirect(redirect_url)

        elif submit == 'Cancel':
            conn.cancel_transaction()
            flash('Entry not deleted')
            return redirect(url_for(
                'table',
                schema=schema,
                table=table,
                subtable=subtable
            ))

    # always cancel transaction
    conn.cancel_transaction()

    if count:
        return render_template(
            'pages/delete.html',
            table_name=table_name,
            message=message.splitlines(),
            restriction=str(_id),
            url=url_for(
                'table',
                schema=schema,
                table=table,
                subtable=subtable
            )
        )
    else:
        flash("Nothing to delete", 'error')
        return redirect(redirect_url)


@app.route('/table/<schema>/<table>',
           defaults={'subtable': None}, methods=['GET', 'POST'])
@app.route('/table/<schema>/<table>/<subtable>', methods=['GET', 'POST'])
@login_required
def table(schema, table, subtable):
    if (  # people can edit their entries but can't create new users
        (request.args.get('edit', None) != "True")
        and ((schema, table)
             == (config.user_table.database, config.user_table.name))
    ):
        return redirect(url_for('register'))
    elif (
        (schema, table)
        == (config.group_table.database, config.group_table.name)
    ):
        return redirect(url_for('registergroup'))
    elif (
        (schema, table)
        == (config.assigned_table.database, config.assigned_table.name)
    ):
        return redirect(url_for('assigngroup'))

    # TODO this should be handled in SQL GRANT permissions
    if (schema, table) == ('subjects', 'FlyStock'):
        override_permissions = True
    else:
        override_permissions = False

    subtable = request.args.get('subtable', subtable)
    edit_url = url_for(
        'table', schema=schema, table=table, subtable=subtable)
    overwrite_url = url_for(
        'table', schema=schema, table=table, subtable=subtable)

    return form_template(
        schema, table, subtable, edit_url, overwrite_url, page='table',
        override_permissions=override_permissions
    )



"""handling of errors
"""

import traceback
from flask import flash, render_template, jsonify
from loris.app.app import app
from loris.errors import LorisError


@app.errorhandler(LorisError)
def loriserror(e):
    flash(
        'This was a Loris internal exception; '
        'read the flashed error message below for information.', 'warning')
    error_traceback = traceback.format_exc().splitlines()
    flash(error_traceback[-1], 'error')
    return render_template('pages/error.html', traceback=error_traceback), 500


@app.errorhandler(Exception)
def error(e):
    error_traceback = traceback.format_exc().splitlines()
    flash(error_traceback[-1], 'error')
    return render_template('pages/error.html', traceback=error_traceback), 500



#!/usr/bin/env python

from setuptools import setup, find_packages
import os
import sys

min_py_version = (3, 7)

if sys.version_info < min_py_version:
    sys.exit(
        'Loris is only supported '
        'on Python {}.{} or higher'.format(*min_py_version))

path = os.path.abspath(os.path.dirname(__file__))

long_description = 'Data management and analysis application for DataJoint'

with open(os.path.join(path, 'requirements.txt')) as f:
    requirements = f.read().split()


setup(
    name='lorisapp',
    version='0.1.4dev1',
    description='Loris',
    long_description=long_description,
    author='Matthias Christenson',
    author_email='gucky@gucky.eu',
    install_requires=requirements,
    keywords='database application',
    packages=find_packages(exclude=["docs", "images", "tests"]),
    include_package_data=True,
    scripts=['bin/lorisapp'],
)



"""views
"""

from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd

from loris import config
from loris.app.app import app
from loris.app.templates import form_template, joined_table_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password, grantprivileges


@app.route('/login', methods=['GET', 'POST'])
def login():
    if current_user.is_authenticated:
        return redirect(url_for('home'))

    form = LoginForm()

    if form.validate_on_submit():
        user = User(form.user_name.data)
        if user.user_name == 'root':
            flash('Cannot login as root', 'error')
        elif not user.user_exists or not user.check_password(form.password.data):
            flash('Invalid username or password', 'error')
        elif not user.is_active:
            flash(f'User {user.user_name} is an inactive member', 'error')
        elif form.password.data == config['standard_password']:
            flash('Please change your password', 'warning')
            login_user(user)
            return redirect(url_for('change'))
        else:
            login_user(user)
            redirect_url = request.args.get('target', None)
            if redirect_url is None:
                return redirect(url_for('home'))
            else:
                return redirect(redirect_url)

    return render_template(
        'pages/login.html',
        form=form,
    )


@app.route("/logout")
@login_required
def logout():
    # user_configs.pop(current_user.user_name, None)
    logout_user()
    flash('Successful logout!')
    return redirect(url_for('login'))


@app.route("/change", methods=['GET', 'POST'])
@login_required
def change():
    """change password
    """

    form = PasswordForm()

    if form.validate_on_submit():

        user = User(current_user.user_name)
        if (
            not user.user_exists
            or not user.check_password(form.old_password.data)
        ):
            flash('Old password incorrect', 'error')
        elif form.old_password.data == form.new_password.data:
            flash('New and old password match', 'error')
        else:
            change_password(current_user.user_name, form.new_password.data)
            flash('Successfully changed password and logged in')
            login_user(user)

            redirect_url = request.args.get('target', None)
            if redirect_url is None:
                return redirect(url_for('home'))
            else:
                return redirect(redirect_url)

    return render_template(
        'pages/change.html',
        form=form
    )


@app.route('/')
@login_required
def home():
    # refresh session
    app.session_refresh()
    return render_template(
        'pages/home.html',
        user=current_user.user_name
    )


@app.route('/about')
@login_required
def about():
    return render_template('pages/about.html')


@app.route('/register', methods=['GET', 'POST'])
@login_required
def register():

    if current_user.user_name not in config['administrators']:
        flash("Only administrators are allowed to register users", "warning")
        return redirect(
            url_for(
                'table',
                schema=config['user_schema'],
                table=config['user_table'],
                edit="True",
                _id={config['user_name']: current_user.user_name}
            )
        )

    user_class = config.user_table

    dynamicform, form = config.get_dynamicform(
        f'{config["user_schema"]}.{config["user_table"]}',
        user_class, DynamicForm
    )

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit == 'Register':
            if form.validate_on_submit():
                try:
                    dynamicform.insert(form)
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    dynamicform.reset()
                    formatted_dict = form.get_formatted()
                    grantuser(
                        formatted_dict[config['user_name']],
                        adduser=True
                    )
                    flash("User created", 'success')

        form.append_hidden_entries()

    edit_url = url_for(
        'table', schema=config['user_schema'], table=config['user_table'])
    delete_url = url_for(
        'delete', schema=config['user_schema'], table=config['user_table'])

    data = dynamicform.get_jsontable(
        edit_url, delete_url,
    )

    return render_template(
        'pages/register.html',
        form=form,
        data=data,
        toggle_off_keys=[0]
    )


@app.route('/registeredusers', methods=['GET', 'POST'])
@login_required
def registeredusers():
    """registeredusers
    """
    delete_url = url_for(
        'delete',
        schema=config['user_schema'],
        table=config['user_table'],
        subtable=None)

    return joined_table_template(
        [config.user_table],
        'Registered Users',
        edit_url=url_for(
            'table',
            schema=config['user_schema'],
            table=config['user_table'],
            subtable=None
        ),
        delete_url=delete_url
    )


@app.route('/emergencycontacts', methods=['GET', 'POST'])
@login_required
def emergencycontacts():
    """emergencycontacts
    """
    if not hasattr(config.user_table, 'EmergencyContact'):
        return redirect(url_for('registeredusers'))

    delete_url = url_for(
        'delete',
        schema=config['user_schema'],
        table=config['user_table'],
        subtable=None)

    return joined_table_template(
        [config.user_table.EmergencyContact],
        'Emergency Contacts',
        edit_url=url_for(
            'table',
            schema=config['user_schema'],
            table=config['user_table'],
            subtable=None
        ),
        delete_url=delete_url
    )


@app.route('/registergroup', methods=['GET', 'POST'])
@login_required
def registergroup():
    """setup a group
    """

    if current_user.user_name not in config['administrators']:
        flash("Only administrators are allowed to register groups", "warning")
        return redirect(url_for('home'))

    group_class = config.group_table

    dynamicform, form = config.get_dynamicform(
        f'{config["group_schema"]}.{config["group_table"]}',
        group_class, DynamicForm
    )

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit == 'Register':
            if form.validate_on_submit():
                try:
                    dynamicform.insert(form)
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    dynamicform.reset()
                    config.create_group_schemas()
                    flash("Project group created", 'success')

        form.append_hidden_entries()

    edit_url = url_for(
        'table', schema=config['group_schema'], table=config['group_table'])
    delete_url = url_for(
        'delete', schema=config['group_schema'], table=config['group_table'])

    data = dynamicform.get_jsontable(
        edit_url, delete_url,
    )

    return render_template(
        'pages/group.html',
        form=form,
        data=data,
        toggle_off_keys=[0]
    )


@app.route('/assigngroup', methods=['GET', 'POST'])
@login_required
def assigngroup():
    """setup a group
    """

    if current_user.user_name not in config['administrators']:
        flash("Only administrators are allowed to assign groups", "warning")
        return redirect(url_for('home'))

    # only assign yourself to a group
    # TODO what should the authorization behavior be?
    kwargs = {
        config['user_name']: current_user.user_name
    }

    if current_user.user_name in config['administrators']:
        readonly = []
    else:
        readonly = [config['user_name']]

    group_class = config.assigned_table

    dynamicform, form = config.get_dynamicform(
        f'{config["assignedgroup_schema"]}.{config["assignedgroup_table"]}',
        group_class, DynamicForm, **kwargs
    )

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit == 'Register':
            if form.validate_on_submit():
                try:
                    formatted_dict = dynamicform.insert(form)
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    dynamicform.reset()
                    config.create_group_schemas()
                    config.refresh_permissions()
                    user = formatted_dict[config['user_name']]
                    group = formatted_dict[config['group_name']]
                    grantprivileges(
                        user,
                        config['connection'],
                        {f'{group}.*': 'ALL PRIVILEGES'}
                    )
                    flash(
                        f"Project {group} now includes user {user}",
                        'success'
                    )

        form.append_hidden_entries()

    delete_url = url_for(
        'delete',
        schema=config["assignedgroup_schema"],
        table=config["assignedgroup_table"])

    data = dynamicform.get_jsontable(
        delete_url=delete_url,
    )

    return render_template(
        'pages/assigngroup.html',
        form=form,
        data=data,
        toggle_off_keys=[0],
        readonly=readonly
    )



@app.route(f"{config['tmp_folder']}/<path:filename>")
@login_required
def tmpfile(filename):
    return send_from_directory(config['tmp_folder'], filename)



"""fixed forms
"""

import pandas as pd
import numpy as np
import os
import glob
import graphviz
from flask_wtf import FlaskForm as Form
from wtforms import Form as NoCsrfForm
from flask import render_template, request, flash, url_for, redirect
import datajoint as dj
from datajoint.table import lookup_class_name
from wtforms import Form as NoCsrfForm
from wtforms import StringField, IntegerField, BooleanField, FloatField, \
    SelectField, FieldList, FormField, HiddenField, TextAreaField, PasswordField
from wtforms.validators import InputRequired, Optional, NumberRange, \
    ValidationError, Length, UUID, URL, Email, EqualTo

from loris import config
from loris.app.forms.formmixin import FormMixin, DynamicFileField, \
    DictField, ListField, JsonSerializableValidator, RestrictionField, \
    OptionalJsonSerializableValidator, Extension, CamelCaseValidator
from loris.errors import LorisError


RESTRICTION_DESCRIPTION = (
    'a sql where clause to apply to the joined '
    'table or a list of dicts or a dict of restrictions '
    '(do not include the WHERE command)'
)

RESTRICTION_LABEL = (
    'restriction - '
    '<a href="https://www.tutorialgateway.org/mysql-where-clause" '
    'target="_blank">help</a>'
)


class LoginForm(Form, FormMixin):
    user_name = StringField(
        'user name',
        description='SQL database username',
        validators=[InputRequired()]
    )
    password = PasswordField(
        'password',
        description='user password',
        validators=[InputRequired()]
    )


class SettingsNameForm(NoCsrfForm, FormMixin):
    settings_name = StringField(
        'settings name',
        description='save configuration under this name',
        validators=[InputRequired()]
    )


class PasswordForm(Form, FormMixin):
    old_password = PasswordField(
        'old password',
        description='old password',
        validators=[InputRequired()]
    )
    new_password = PasswordField(
        'new password',
        description='new password',
        validators=[InputRequired()]
    )
    repeat_password = PasswordField(
        'repeat password',
        description='repeat password',
        validators=[
            Length(min=10),
            InputRequired(),
            EqualTo('new_password', message='Passwords must match')
        ],
    )


class PartTableCreationForm(NoCsrfForm, FormMixin):
    table_name = StringField(
        'part table name',
        description='camel-case name of part table',
        validators=[
            InputRequired(),
            CamelCaseValidator()
        ]
    )
    definition = TextAreaField(
        'definition',
        description='datajoint definition of a part table',
        validators=[
            InputRequired(),
        ]
    )


def dynamic_autoscriptform():

    from loris.app.app import app

    tables = [
        f"{schema}.{name}"
        for schema, names in app.config['tables'].items()
        for name in names
    ]

    class AutoscriptForm(Form, FormMixin):
        autoscript = SelectField(
            'autoscript',
            description='autoscript',
            choices=[
                (ele, os.path.split(ele)[-1])
                for ele in
                glob.glob(os.path.join(config['autoscript_folder'], '*'))
                if os.path.isdir(ele)
                and not os.path.basename(ele).startswith('_')
            ],
            validators=[InputRequired()]
        )
        experiment = SelectField(
            'experiment',
            description='experiment',
            choices=[
                (ele, ele)
                for ele in
                tables
            ]
        )

    return AutoscriptForm


def dynamic_tablecreationform(user_name):

    class TableCreationForm(Form, FormMixin):
        schema = SelectField(
            'schema',
            description='schema',
            choices=[(key, key) for key in config.schemas_of_user(user_name)],
            default=user_name
        )
        table_name = StringField(
            'table name',
            description='camel-case name of table',
            validators=[
                InputRequired(),
                CamelCaseValidator()
            ]
        )
        table_type = SelectField(
            'table type',
            description='type of table',
            validators=[
                InputRequired()
            ],
            choices=[
                (ele, ele)
                for ele in [
                    'Manual', 'Imported', 'Computed',
                    'AutoImported', 'AutoComputed'
                ]
            ],
            default='Manual'
        )
        definition = TextAreaField(
            (
                'definition - '
                '<a href="https://docs.datajoint.io/python/definition'
                '/04-Definition-Syntax.html" target="_blank">help</a>'
            ),
            description=('datajoint definition of table'),
            validators=[
                InputRequired(),
            ]
        )
        part_tables = FieldList(
            FormField(PartTableCreationForm),
            min_entries=1
        )

        @staticmethod
        def dynamic_table_class(table_name, table_type, definition_string):
            """dynamically create datajoint table
            """

            class TableClass(getattr(dj, table_type)):
                definition = definition_string
                name = table_name

            return TableClass

        def declare_table(self):
            """Create table
            """
            formatted_dict = self.get_formatted()

            schema = config['schemata'][formatted_dict['schema']].schema

            table_class = self.dynamic_table_class(
                formatted_dict['table_name'],
                formatted_dict['table_type'],
                formatted_dict['definition'])

            if formatted_dict['part_tables'] is not None:
                for part_table in formatted_dict['part_tables']:
                    part_table_class = self.dynamic_table_class(
                        part_table['table_name'],
                        'Part',
                        part_table['definition']
                    )

                    setattr(
                        table_class,
                        part_table['table_name'],
                        part_table_class
                    )

            # get complete context and declare table
            context = {
                **config['schemata'],
                **(config['schemata'][formatted_dict['schema']].__dict__
                   if schema.context is None else schema.context)
            }

            setattr(
                config['schemata'][formatted_dict['schema']],
                formatted_dict['table_name'],
                schema(table_class, context=context)
            )

    return TableCreationForm


def dynamic_jointablesform():

    class JoinTablesForm(Form, FormMixin):
        tables_dict = config['tables']
        tables = FieldList(
            SelectField(
                'tables',
                description='tables',
                choices=[(key, key) for key in tables_dict],
                validators=[InputRequired()]
            ),
            min_entries=1,
        )
        restriction = RestrictionField(
            RESTRICTION_LABEL,
            description=RESTRICTION_DESCRIPTION,
            validators=[
                Optional(),
                OptionalJsonSerializableValidator(is_instance=(dict, list))
            ],
            render_kw={
                'nullable': True
            }
        )

    return JoinTablesForm


class ModuleForm(NoCsrfForm, FormMixin):
    # TODO implement validators
    python_file = DynamicFileField(
        'python file',
        description='python file to upload with function',
        validators=[Optional(), Extension(['py'])]
    )
    python_module = StringField(
        'python module',
        description='name of python module, if no file is provided',
        validators=[Optional()],
        render_kw={
            'nullable': True
        }
    )

    def get_formatted(self):

        formatted_dict = super().get_formatted()

        if formatted_dict['python_file'] is None and formatted_dict['python_module'] is None:
            raise LorisError('No python file or module was given.')
        elif formatted_dict['python_file']:
            return formatted_dict['python_file']
        else:
            return formatted_dict['python_module']


class FuncForm(NoCsrfForm, FormMixin):
    module = FormField(ModuleForm)
    function = StringField(
        'function',
        description='function name in the module',
        validators=[InputRequired()]
    )
    args = ListField(
        'args',
        description='list of arguments for init if function is class',
        validators=[Optional(), JsonSerializableValidator(list)],
        render_kw={
            'nullable': True
        }
    )
    kwargs = DictField(
        'kwargs',
        description='dict of keyword arguments for init if function is class',
        validators=[Optional(), JsonSerializableValidator(dict)],
        render_kw={
            'nullable': True
        }
    )

    def get_formatted(self):

        formatted = super().get_formatted()

        if formatted['args'] is None and formatted['kwargs'] is None:
            return (
                formatted['module'],
                formatted['function']
            )
        else:
            return (
                formatted['module'],
                formatted['function'],
                ([] if formatted['args'] is None else formatted['args']),
                ({} if formatted['kwargs'] is None else formatted['kwargs'])
            )


def dynamic_settingstableform(table_class):

    class FetchTableForm(NoCsrfForm, FormMixin):
        table_name = SelectField(
            'table name',
            description='choose table',
            choices=[(table_name, table_name) for table_name in config['tables']],
            validators=[InputRequired()]
        )
        proj_list = ListField(
            'projections',
            description='arguments to project',
            validators=[Optional(), JsonSerializableValidator(list)],
            render_kw={
                'nullable': True
            }
        )
        proj_dict = DictField(
            'renamed projections',
            description='arguments to project and rename',
            validators=[Optional(), JsonSerializableValidator(dict)],
            render_kw={
                'nullable': True
            }
        )

        def get_formatted(self):

            formatted = super().get_formatted()

            return {
                formatted['table_name'] : (
                    ([] if formatted['proj_list'] is None else formatted['proj_list']),
                    ({} if formatted['proj_dict'] is None else formatted['proj_dict'])
                )
            }

    class SettingstableForm(Form, FormMixin):
        settings_name = StringField(
            'settings name',
            description='unique name for settings used to autopopulate',
            validators=[InputRequired(), Length(max=63)]
        )
        description = TextAreaField(
            'description',
            description='longer description to describe settings',
            validators=[Optional(), Length(max=4000)],
            render_kw={
                'nullable': True
            }
        )
        func = FormField(FuncForm)
        global_settings = DictField(
            'global settings',
            description='dict of keyword arguments to pass to function for every entry',
            validators=[InputRequired(), JsonSerializableValidator(dict)]
        )
        entry_settings = DictField(
            'entry settings',
            description='dict of keyword arguments to pass to function specific to each entry as defined by columns of the joined table',
            validators=[InputRequired(), JsonSerializableValidator(dict)]
        )
        fetch_method = SelectField(
            'fetch method',
            description='method used to fetch data',
            choices=[('fetch1', 'fetch1'), ('fetch', 'fetch')],
        )
        fetch_tables = FieldList(
            FormField(FetchTableForm),
            label='fetch tables',
            min_entries=1,
            render_kw={
                'nullable': True
            }
        )
        assign_output = SelectField(
            'assign output',
            description='assign the output of the function to a single column',
            choices=[('NULL', 'NULL')]+[
                (str(ele), str(ele))
                for ele in table_class.child_table().heading.secondary_attributes
            ]
        )
        restrictions = RestrictionField(
            RESTRICTION_LABEL,
            description=RESTRICTION_DESCRIPTION,
            validators=[Optional(), OptionalJsonSerializableValidator((list, dict))],
            render_kw={
                'nullable': True
            }
        )
        parse_unique = ListField(
            'parse as unique',
            description='list of unique entries when using fetch - not wrapped into numpy.array',
            validators=[Optional(), JsonSerializableValidator(list)],
            render_kw={
                'nullable': True
            }
        )

        def get_formatted(self):

            formatted = super().get_formatted()

            if formatted['fetch_tables'] is not None:
                formatted['fetch_tables'] = {
                    key: value
                    for table_dict in formatted['fetch_tables']
                    for key, value in table_dict.items()
                }

            return formatted

    return SettingstableForm


def dynamic_runform(table_class):

    class RunForm(Form, FormMixin):
        settings_name = SelectField(
            'settings method',
            description='settings name of function to apply',
            validators=[InputRequired()],
            choices=[
                (ele, ele)
                for ele
                in table_class.settings_table.proj().fetch()[
                    table_class.settings_table.primary_key[0]
                ]
            ]
        )
        restriction = RestrictionField(
            RESTRICTION_LABEL,
            description=RESTRICTION_DESCRIPTION,
            validators=[InputRequired(), OptionalJsonSerializableValidator((list, dict))],
        )
        limit = IntegerField(
            'autopopulate limit',
            description='limit of entries to autopopulate',
            default=1,
            validators=[Optional(), NumberRange(min=1)]
        )
        suppress_errors = BooleanField(
            'suppress errors',
            description='do not terminate execution',
            default=False,
            validators=[Optional()]
        )
        multiprocess = IntegerField(
            'multiprocess',
            description='use multiple cpus - 0 means False',
            default=0,
            validators=[
                InputRequired(),
                NumberRange(min=0, max=config['max_cpu'])
            ]
        )

    return RunForm



"""utility functions for database
"""

import os

import loris


class Placeholder:
    """a file placeholder class for the placeholder custom attribute
    """

    filename = "placeholder.txt"

    def __init__(self, value):
        self.value = value

    @classmethod
    def read(cls, filepath):
        """read value
        """
        filename = os.path.split(filepath)[-1]
        if filename == cls.filename:
            with open(filepath, 'rb') as f:
                obj = f.read()
            return cls(obj)

        return filepath

    def write(self):
        """write value
        """
        if not os.path.exists(self.value):

            filepath = os.path.join(loris.config['tmp_folder'], self.filename)
            with open(filepath, 'wb') as f:
                f.write(self.value)
            return filepath

        return self.value


class ProcessMixin:
    """a mixin to process strings or other objects for
    blob serialization or attaching.
    """

    def put_process(obj):
        if obj is None:
            return
        return obj

    def get_process(value):

        return value



"""
"""

from flask import g
from werkzeug.local import LocalProxy

from loris import config
from loris.app.wiki.wiki import Wiki


def get_wiki():
    wiki = getattr(g, '_wiki', None)
    if wiki is None:
        wiki = g._wiki = Wiki(config['wiki_folder'])
    return wiki


current_wiki = LocalProxy(get_wiki)



"""wiki class
"""

import os
import re

from flask import url_for, redirect

from loris.app.wiki.page import Page
from loris.app.wiki.utils import clean_url


class Wiki(object):
    def __init__(self, root):
        self.root = root

    def path(self, url):
        return os.path.join(self.root, url + '.md')

    def exists(self, url):
        path = self.path(url)
        return os.path.exists(path)

    def get(self, url):
        path = os.path.join(self.root, url + '.md')
        if self.exists(url):
            return Page(path, url)
        return None

    def get_bare(self, url):
        path = self.path(url)
        if self.exists(url):
            return False
        return Page(path, url, new=True)

    def move(self, url, newurl):
        source = os.path.join(self.root, url) + '.md'
        target = os.path.join(self.root, newurl) + '.md'
        # normalize root path (just in case somebody defined it absolute,
        # having some '../' inside) to correctly compare it to the target
        root = os.path.normpath(self.root)
        # get root path longest common prefix with normalized target path
        common = os.path.commonprefix((root, os.path.normpath(target)))
        # common prefix length must be at least as root length is
        # otherwise there are probably some '..' links in target path leading
        # us outside defined root directory
        if len(common) < len(root):
            raise RuntimeError(
                'Possible write attempt outside content directory: '
                '%s' % newurl)
        # create folder if it does not exists yet
        folder = os.path.dirname(target)
        if not os.path.exists(folder):
            os.makedirs(folder)
        os.rename(source, target)

    def delete(self, url):
        path = self.path(url)
        if not self.exists(url):
            return False
        os.remove(path)
        return True

    def index(self):
        """
            Builds up a list of all the available pages.

            :returns: a list of all the wiki pages
            :rtype: list
        """
        # make sure we always have the absolute path for fixing the
        # walk path
        pages = []
        root = os.path.abspath(self.root)
        for cur_dir, _, files in os.walk(root):
            # get the url of the current directory
            cur_dir_url = cur_dir[len(root)+1:]
            for cur_file in files:
                path = os.path.join(cur_dir, cur_file)
                if cur_file.endswith('.md'):
                    url = clean_url(os.path.join(cur_dir_url, cur_file[:-3]))
                    page = Page(path, url)
                    pages.append(page)
        return sorted(pages, key=lambda x: x.title.lower())

    def index_dict(self):
        """
        Builds nested directionary of all available pages
        """

        index_dict = {}

        for page in self.index():
            page = page.url.split('/')
            current_dict = index_dict
            for subpage in page:
                if subpage not in current_dict:
                    current_dict[subpage] = {}
                current_dict = current_dict[subpage]

        return index_dict

    def index_by(self, key):
        """
            Get an index based on the given key.

            Will use the metadata value of the given key to group
            the existing pages.

            :param str key: the attribute to group the index on.

            :returns: Will return a dictionary where each entry holds
                a list of pages that share the given attribute.
            :rtype: dict
        """
        pages = {}
        for page in self.index():
            value = getattr(page, key)
            pre = pages.get(value, [])
            pages[value] = pre.append(page)
        return pages

    def get_by_title(self, title):
        pages = self.index(attr='title')
        return pages.get(title)

    def get_tags(self):
        pages = self.index()
        tags = {}
        for page in pages:
            pagetags = page.tags.split(',')
            for tag in pagetags:
                tag = tag.strip()
                if tag == '':
                    continue
                elif tags.get(tag):
                    tags[tag].append(page)
                else:
                    tags[tag] = [page]
        return tags

    def index_by_tag(self, tag):
        pages = self.index()
        tagged = []
        for page in pages:
            if tag in page.tags:
                tagged.append(page)
        return sorted(tagged, key=lambda x: x.title.lower())

    def search(self, term, ignore_case=True, attrs=['title', 'tags', 'body']):
        pages = self.index()
        regex = re.compile(term, re.IGNORECASE if ignore_case else 0)
        matched = []
        for page in pages:
            for attr in attrs:
                if regex.search(getattr(page, attr)):
                    matched.append(page)
                    break
        return matched



"""Wtf Forms
"""

import os
import numpy as np
import json
import re
import warnings

from wtforms import FieldList, FormField, BooleanField, StringField, \
    TextAreaField, SelectField
from wtforms.widgets import HiddenInput
from wtforms import Form as NoCsrfForm
from flask_wtf.file import FileField
from werkzeug.utils import secure_filename
from wtforms.validators import InputRequired, Optional, NumberRange, \
    ValidationError, Length, UUID, URL, Email, StopValidation
from wtforms.widgets import Select
from wtforms.widgets import html_params
from markupsafe import Markup, escape
from wtforms.compat import text_type

from loris import config
from loris.app.forms import NONES


class HtmlLabelSelectWidget(Select):

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        if value is True:
            # Handle the special case of a 'True' value.
            value = text_type(value)

        if isinstance(label, (tuple, list)):
            comment = "(" + (
                label[1]
                if len(label[1]) < 52
                else label[1][:50] + '...'
            ) + ")"
            label = (
                f'{escape(label[0])}  '
                f'&emsp;<small>{escape(comment)}</small>')
            options = dict(kwargs, value=value, title=comment)
        else:
            label = escape(label)
            options = dict(kwargs, value=value)
        if selected:
            options["selected"] = True
        return Markup(
            f"<option {html_params(**options)}>{label}</option>"
        )


class MetaHiddenField(BooleanField):
    widget = HiddenInput()


class TagListField(StringField):
    """Stringfield for a list of separated tags"""

    def __init__(
        self, label='', validators=None, remove_duplicates=True,
        to_lowercase=True, separator=',', **kwargs
    ):
        """
        Construct a new field.

        Parameters
        ----------
        label: The label of the field.
        validators: A sequence of validators to call when validate is called.
        remove_duplicates: Remove duplicates in a case insensitive manner.
        to_lowercase: Cast all values to lowercase.
        separator: The separator that splits the individual tags.
        """
        super(TagListField, self).__init__(label, validators, **kwargs)
        self.remove_duplicates = remove_duplicates
        self.to_lowercase = to_lowercase
        self.separator = separator
        self.data = []

    def _value(self):
        if self.data:
            return u'{} '.format(self.separator).join(self.data)
        else:
            return u''

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = [
                x.strip()
                for x in valuelist[0].split(self.separator)
                if x.strip()
            ]
            if self.remove_duplicates:
                self.data = list(self._remove_duplicates(self.data))
            if self.to_lowercase:
                self.data = [x.lower() for x in self.data]
            if not self.data:
                self.data = None

    @classmethod
    def _remove_duplicates(cls, seq):
        """
        Remove duplicates in a case insensitive,
        but case preserving manner
        """
        d = {}
        for item in seq:
            if item.lower() not in d:
                d[item.lower()] = True
                yield item


class TagDictField(StringField):
    """Stringfield for a list of separated tags"""

    def __init__(
        self, label='', validators=None,
        key_separator='=', separator=',', **kwargs
    ):
        """
        Construct a new field.

        Parameters
        ----------
        label: The label of the field.
        validators: A sequence of validators to call when validate is called.
        key_separator : The separator between key and value.
        separator: The separator that splits the individual tags.
        """
        super(TagDictField, self).__init__(label, validators, **kwargs)
        self.key_separator = key_separator
        self.separator = separator
        self.data = dict()

    def _value(self):
        if self.data:
            return u'{} '.format(self.separator).join([
                f"{key}{self.key_separator}{value}"
                for key, value in self.data.items()
            ])
        else:
            return u''

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = dict([
                [ele.strip() for ele in x.strip().split(self.key_separator)]
                for x in valuelist[0].split(self.separator)
                if x.strip() and (self.key_separator in x.strip())
            ])
            if not self.data:
                self.data = None


class EvalJsonField:

    startswith = None
    endswith = None

    def _value(self):
        if self.data not in NONES:
            return json.dumps(self.data)
        else:
            return ''

    def process_formdata(self, valuelist):
        if valuelist:
            data = valuelist[0]
            if data in NONES:
                self.data = None
            else:
                if (
                    not data.startswith(self.startswith)
                    or not data.endswith(self.endswith)
                ):
                    data = f'{self.startswith}{data}{self.endswith}'
                self.data = json.loads(data)


class ListField(EvalJsonField, StringField):
    startswith = '['
    endswith = ']'


class DictField(EvalJsonField, StringField):
    startswith = '{'
    endswith = '}'


class DictTextArea(EvalJsonField, TextAreaField):
    startswith = '{'
    endswith = '}'


class RestrictionField(StringField):

    def evaluate(self, data):
        if data in NONES:
            return False
        elif data.startswith('['):
            return True
        elif data.startswith('{'):
            return True
        else:
            return False

    def _value(self):
        if self.data in NONES:
            return ''
        elif isinstance(self.data, str):
            return self.data
        else:
            return json.dumps(self.data)

    def process_formdata(self, valuelist):
        if valuelist:
            data = valuelist[0]
            if data in NONES:
                self.data = None
            elif self.evaluate(data):
                self.data = json.loads(data)
            else:
                self.data = data


class DynamicFileField(FileField):
    pass  # TODO


class CamelCaseValidator:

    def __init__(self, name='table name'):

        self.name = name

    def __call__(self, form, field):

        if not re.match(r'[A-Z][a-zA-Z0-9]*', field.data):
            raise ValidationError(
                f'{self.name} must be alphanumeric in CamelCase, '
                'begin with a capital letter.'
            )


class Extension:
    """Extension Validator
    """

    def __init__(self, ext=config['extensions']):
        self.ext = ext

    def __call__(self, form, field):
        filename = field.data.filename
        if not filename:
            return
        extension = os.path.splitext(filename)[-1].strip('.')
        if extension.lower() not in self.ext:
            raise ValidationError(
                f"File {filename} is not of extension: {self.ext}, "
                f"but extension {extension}."
            )


class FilePath:

    def __call__(self, form, field):

        data = field.data
        if not os.path.exists(data):
            raise ValidationError(
                f'Filepath {data} does not exist.'
            )


class LookupNameValidator:

    def __call__(self, form, field):

        data = field.data

        if data in NONES:
            return

        if isinstance(data, str):
            data = data.strip().lower()
            if not data.isidentifier():
                raise ValidationError(
                    f"lookup name '{data}' is not an identifier; "
                    "it contains characters besides alphanumeric and/or "
                    "an underscore."
                )
        else:
            raise ValidationError(
                'Lookup Name must be of string type.'
            )


class ParentInputRequired(Optional):

    def __call__(self, form, field):
        """only required if <new> set
        """

        try:
            existing = getattr(form, 'existing_entries').data
        except AttributeError:
            existing = '<new>'

        if existing == '<new>':
            if field.data in NONES:
                raise ValidationError('Data missing!')

        super().__call__(form, field)


class Always:

    def __call__(self, form, field):
        try:
            existing = getattr(form, 'existing_entries').data
        except AttributeError:
            existing = '<new>'

        if existing == '<new>':
            raise ValidationError("Data completely missing!")


class ParentValidator:

    def __init__(self, primary_key):
        self.primary_key = primary_key

    def __call__(self, form, field):
        """
        """

        data = tuple([getattr(form, key).data for key in self.primary_key])

        if field.data == '<new>':
            if any(idata in NONES for idata in data):
                raise ValidationError(
                    'Must specify new foreign primary key '
                    'if <add new entry> is selected.'
                )
            if str(data) in [ele for ele, ele in field.choices]:
                raise ValidationError(
                    f'Entry {data} already exists in parent table.'
                )

        else:
            if any(idata not in NONES for idata in data):
                raise ValidationError(
                    'If specifying new foreign primary key '
                    'need to set select field to <add new entry>'
                )


class PutValidator:

    def __init__(self, put_test=None):
        self.put_test = put_test

    def __call__(self, form, field):

        data = field.data

        if self.put_test is not None:
            try:
                self.put_test(data)
            except Exception as e:
                raise ValidationError(
                    f'Data did not pass put test: {e}'
                )


class JsonSerializableValidator:

    def __init__(self, is_instance=None, put_test=None):
        self.is_instance = is_instance
        self.put_test = put_test

    def __call__(self, form, field):

        data = field.data

        if data in NONES:
            return

        if not isinstance(data, str):
            if (
                self.is_instance is not None
                and
                not isinstance(data, self.is_instance)
            ):
                raise ValidationError(
                    f'Json Data is not correct type {self.is_instance}, '
                    f'but is {type(data)}'
                )
        else:
            try:
                data = json.loads(data)
                if self.is_instance is not None:
                    assert isinstance(data, self.is_instance), (
                        f'Json Data is not correct type {self.is_instance}, '
                        f'but is {type(data)}'
                    )
            except Exception as e:
                raise ValidationError(
                    f'Data is not a json-serializable: {e}'
                )

        if self.put_test is not None:
            try:
                self.put_test(data)
            except Exception as e:
                raise ValidationError(
                    f'Data did not pass put test: {e}'
                )


class OptionalJsonSerializableValidator(JsonSerializableValidator):

    def __call__(self, form, field):
        if not isinstance(field.data, str):
            super().__call__(form, field)


class BlobFileField(DynamicFileField):
    pass


class AttachFileField(DynamicFileField):
    pass


class FormMixin:
    hidden_entries = None

    def populate_form(self, formatted_dict):
        """populate form from formatted_dict

        Parameters
        ----------
        formatted_dict : dict
            dictionary as formatted by get_formatted method.
        readonly : iterable
            field ids that are readonly.
        """
        if not isinstance(formatted_dict, dict):
            warnings.warn(
                "Formatted dictionary is not of type `dict`, "
                f"but of type `{type(formatted_dict)}`"
            )
            return self

        for field in self:
            key = field.short_name

            if key in formatted_dict:
                data = formatted_dict[key]

                if data is None:
                    continue

                elif isinstance(field, FieldList):
                    while len(field) > 1:
                        field.pop_entry()

                    subfield = field[0]

                    if isinstance(subfield, FormField):
                        for idata in data:
                            sanitized_data = subfield.populate_form(idata).data
                            field.append_entry(sanitized_data)

                    else:
                        for idata in data:
                            field.append_entry(idata)

                elif isinstance(field, FormField):
                    field.populate_form(data)

                else:
                    field.process_data(data)

        return self

    def rm_hidden_entries(self):
        """removes hidden entries in form to allow for dynamic lists
        """

        self.hidden_entries = {}
        self.subhidden_entries = []

        for field in self:

            if isinstance(field, FormField):
                field.rm_hidden_entries()
                self.subhidden_entries.append(field)

            elif isinstance(field, FieldList):

                hidden_entry = field.entries.pop(0)
                self.hidden_entries[field] = hidden_entry

                for subfield in field:
                    # check if subfield is formfield and pop entry
                    if isinstance(subfield, FormField):
                        subfield.rm_hidden_entries()
                        self.subhidden_entries.append(subfield)

    def append_hidden_entries(self):
        """reinsert hidden entries in form to allow for dynamiv lists
        """

        if self.hidden_entries is None:
            pass

        else:
            for subfield in self.subhidden_entries:
                subfield.append_hidden_entries()

            for field, hidden_entry in self.hidden_entries.items():

                field.entries.insert(0, hidden_entry)

            self.hidden_entries = None

    @staticmethod
    def get_field_data(field):

        def _get_field_data(field, nan_return):

            if field.data in NONES:
                return nan_return

            if isinstance(field, FileField):
                filename = secure_filename(field.data.filename)
                if filename in NONES:
                    return nan_return
                filepath = os.path.join(config['tmp_folder'], filename)
                field.data.save(filepath)
                return filepath

            return field.data

        if isinstance(field, BooleanField):
            return _get_field_data(field, False)
        else:
            return _get_field_data(field, None)

    def get_formatted(self):
        """get data from form and reformat for
        execution and saving
        """
        formatted_dict = {}
        for field in self:

            key = field.short_name

            if key == 'csrf_token':
                continue

            elif isinstance(field, FieldList):
                formatted_dict[key] = []

                for subfield in field:
                    if isinstance(subfield, FormField):
                        formatted_dict[key].append(subfield.get_formatted())

                    else:
                        formatted_dict[key].append(
                            self.get_field_data(subfield))

                # No Nones are allowed in a list, they will be removed
                # or the key will be set to None
                are_None = [value is None for value in formatted_dict[key]]

                if np.all(are_None):
                    formatted_dict[key] = None

                else:
                    formatted_dict[key] = [
                        value
                        for value in formatted_dict[key]
                        if value is not None]

            elif isinstance(field, FormField):
                formatted_dict[key] = field.get_formatted()

            else:
                formatted_dict[key] = self.get_field_data(field)
        return formatted_dict


class ManualLookupForm(NoCsrfForm, FormMixin):
    """parent class for manual lookup forms for instance checking.
    """
    pass


class ParentFormField(FormField):
    """FormField for a parent class for instance checking.
    """
    pass



"""Customized attributes
"""

import re
import os
import shutil
import json
from numbers import Number, Integral

import numpy as np
import datajoint as dj

from loris.database.attribute_mixin import Placeholder, ProcessMixin
from loris.errors import LorisSerializationError
from loris.io import spickledumps, spickleloads


class TrueBool(dj.AttributeAdapter):

    attribute_type = 'bool'

    def put(self, obj):
        if obj is None or np.isnan(obj):
            return
        return bool(obj)

    def get(self, value):
        return bool(value)


class SavePickleMixin:

    def put(self, obj):
        if obj is None:
            return
        return spickledumps(obj)

    def get(self, value):
        return spickleloads(value)


class PickleBlob(SavePickleMixin, dj.AttributeAdapter):

    attribute_type = 'longblob'


class PickleData(SavePickleMixin, dj.AttributeAdapter):

    attribute_type = 'blob@datastore'


class ArrMixin:

    def __init__(self, ndim=None, shape=None, atleast=False):
        self.ndim = ndim
        self.shape = shape
        self.atleast = atleast

    def put(self, obj):
        if obj is None:
            return

        if not isinstance(obj, np.ndarray):
            raise TypeError(f"object must be of type `numpy.ndarray`, but is of type `{type(obj)}`")

        if self.ndim is None and self.shape is None:
            return obj

        if self.ndim is not None:
            if self.atleast:
                assert obj.ndim >= self.ndim, f"object has dimensionality `{obj.ndim}`, but only dimensionality `{self.ndim}` or above is allowed."
            else:
                assert obj.ndim == self.ndim, f"object has dimensionality `{obj.ndim}`, but only dimensionality `{self.ndim}` is allowed."

        if self.shape is not None:
            for dim, (allowed_len, is_len) in enumerate(zip(self.shape[::-1], obj.shape[::-1])):
                assert allowed_len == is_len, f"object has length `{is_len}` for dimension {dim}, but must be length `{allowed_len}`"

        return obj

    def get(self, obj):
        return obj


class ArrStore(ArrMixin, dj.AttributeAdapter):
    attribute_type = 'blob@datastore'


class ArrBlob(ArrMixin, dj.AttributeAdapter):
    attribute_type = 'longblob'


class LookupName(dj.AttributeAdapter):
    """a string with stripped of
    """

    attribute_type = 'varchar(127)'

    def put(self, obj):
        if obj is None:
            return

        if isinstance(obj, str):
            obj = obj.strip()  # .lower() - removed since MySQL does not distinguish
        else:
            raise dj.DataJointError(
                f"lookup name '{obj}' must be of type "
                f"'str' and not '{type(obj)}'."
            )

        if not obj.isidentifier():
            raise dj.DataJointError(
                f"lookup name '{obj}' is not an identifier; "
                "it containes characters besides alphanumeric and/or "
                "an underscore."
            )

        return obj

    def get(self, value):
        return value


class PrefixId(dj.AttributeAdapter):
    """
    Prefix id
    """

    attribute_type = 'varchar(127)'

    def __init__(self, prefix=''):
        super().__init__()
        self.prefix = prefix

    def put(self, obj):
        if obj is None:
            return

        if isinstance(obj, str):
            if obj.startswith(self.prefix):
                obj = obj[len(self.prefix):]
            if obj.isnumeric():
                obj = int(obj)
            else:
                raise dj.DataJointError(
                    "Prefix Id must be an integer, but string is `{obj}`."
                )
        elif isinstance(obj, Integral):
            pass
        else:
            raise dj.DataJointError(
                f"Prefix Id must be an integer; but is of type `{type(obj)}`."
            )

        return f"{self.prefix}{obj}"

    def get(self, value):
        return value


class ListString(dj.AttributeAdapter):

    attribute_type = 'varchar(4000)'

    def __init__(self, ele_truth_call=None, error_message=None):
        self.ele_truth_call = ele_truth_call
        self.error_message = error_message

    def put(self, obj):
        if obj is None:
            return
        if isinstance(obj, str):
            try:
                obj = json.loads(obj)
            except Exception as e:
                raise LorisSerializationError(
                    f'Data is not a json-serializable: {e}'
                )
        assert isinstance(obj, (list, tuple)), \
            f'object must be list or tuple for liststring type: {type(obj)}'
        if self.ele_truth_call is not None:
            assert all([self.ele_truth_call(ele) for ele in obj]), \
                (
                    f'Not all object elements pass truth call.'
                    if self.error_message is None else self.error_message
                )
        return json.dumps(obj)

    def get(self, value):
        return json.loads(value)


class DictString(dj.AttributeAdapter):

    attribute_type = 'varchar(4000)'

    def __init__(self, val_truth_call=None, error_message=None):
        self.val_truth_call = val_truth_call
        self.error_message = error_message

    def put(self, obj):
        if obj is None:
            return
        if isinstance(obj, str):
            try:
                obj = json.loads(obj)
            except Exception as e:
                raise LorisSerializationError(
                    f'Data is not a json-serializable: {e}'
                )
        assert isinstance(obj, (dict)), \
            f'object must be dict for dictstring type: {type(obj)}'
        if self.val_truth_call is not None:
            assert all([self.val_truth_call(val) for val in obj.values()]), \
                (
                    f'Not all object values pass truth call.'
                    if self.error_message is None else self.error_message
                )
        return json.dumps(obj)

    def get(self, value):
        return json.loads(value)


class Chromosome(dj.AttributeAdapter):

    attribute_type = 'varchar(511)'

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        assert isinstance(obj, str), (
            f"object is not of type string, "
            f"but {type(obj)} for chromosome attribute")

        obj = obj.strip()

        return obj

    def get(self, value):
        return value


class Email(dj.AttributeAdapter):

    attribute_type = 'varchar(255)'

    @staticmethod
    def is_email(obj):

        regex = r"^\w+([\.-]?\w+)*@\w+([\.-]?\w+)*(\.\w{2,3})+$"
        return re.fullmatch(regex, obj) is not None

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        assert isinstance(obj, str), (
            f"object is not of type string, "
            f"but {type(obj)} for email attribute")

        obj = obj.strip()

        if not self.is_email(obj):
            raise dj.DatajointError(
                f"string {obj} is not a valid email for attribute {self}"
            )

        return obj

    def get(self, value):
        return value


class Link(dj.AttributeAdapter):

    attribute_type = 'varchar(511)'

    @staticmethod
    def is_url(obj):

        regex = re.compile(
            r'^(?:http|ftp)s?://' # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
            r'localhost|' #localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
            r'(?::\d+)?' # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE
        )

        return re.fullmatch(regex, obj) is not None

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        assert isinstance(obj, str), (
            f"object is not of type string, "
            f"but {type(obj)} for link attribute")

        obj = obj.strip()

        if not self.is_url(obj):
            raise dj.DatajointError(
                f"string {obj} is not a url for attribute {self}"
            )

        return obj

    def get(self, value):
        return value


class FlyIdentifier(dj.AttributeAdapter):

    attribute_type = 'varchar(255)'

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        assert isinstance(obj, str), (
            f"object is not of type string, "
            f"but {type(obj)} for fly identifier attribute")

        obj = obj.strip()

        return obj

    def get(self, value):
        return value


class Phone(dj.AttributeAdapter):

    attribute_type = 'varchar(16)'

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        assert isinstance(obj, str), (
            f"object is not of type string, "
            f"but {type(obj)} for phone attribute")

        obj = obj.strip()

        return obj

    def get(self, value):
        return value


class CrossSchema(dj.AttributeAdapter):

    attribute_type = 'attach@attachstore'

    def put(self, obj):
        """perform checks before putting
        """

        if obj is None:
            return

        return obj

    def get(self, value):
        return value


class TarFolder(dj.AttributeAdapter):

    attribute_type = 'attach@attachstore'

    def put(self, obj):
        """perform checks before putting and archive folder
        """

        if obj is None:
            return

        assert os.path.exists(obj), f'path {obj} does not exist.'

        return shutil.make_archive(obj, 'tar', obj)

    def get(self, value):
        """unpack zip file
        """
        unpacked_file = os.path.splitext(value)[0]
        shutil.unpack_archive(value, unpacked_file)
        return unpacked_file


class AttachProcess(dj.AttributeAdapter, ProcessMixin):

    attribute_type = 'attach@attachstore'

    def put(self, obj):
        return self.put_process(obj)

    def get(self, value):
        return self.get_process(value)


class AttachPlaceholder(dj.AttributeAdapter, ProcessMixin):

    attribute_type = 'attach@attachstore'

    def put(self, obj):
        """perform checks before putting and archive folder
        """

        if obj is None:
            return

        obj = self.put_process(obj)
        return Placeholder(obj).write()

    def get(self, value):
        """get file
        """
        return self.get_process(Placeholder.read(value))


class FolderPath(dj.AttributeAdapter):

    attribute_type = 'varchar(511)'

    def put(self, obj):

        if obj is None:
            return
        elif not isinstance(obj, str):
            raise TypeError(
                f"Object must be string type, but is {type(obj)}"
            )

        if not os.path.exists(obj):
            raise ValueError(
                f"Path `{obj}` does not exist."
            )

        if not os.path.isdir(obj):
            raise ValueError(
                f"Path `{obj}` is not a directory."
            )

        return obj

    def get(self, value):
        return value


class DreyeJSON(dj.AttributeAdapter):
    """
    Use dreye (de)serializer
    """

    attribute_type = 'blob@datastore'

    def put(self, obj):

        if obj is None:
            return

        from dreye.io.serialization import dump_json
        return dump_json(obj)

    def get(self, value):
        from dreye.io.serialization import load_json
        return load_json(value)


chr = Chromosome()
link = Link()
flyidentifier = FlyIdentifier()
crossschema = CrossSchema()
truebool = TrueBool()
tarfolder = TarFolder()
liststring = ListString()
dreyejson = DreyeJSON()
tags = ListString(
    lambda x: isinstance(x, str),
    'Elements of list must be strings.'
)
numberlist = ListString(
    lambda x: isinstance(x, Number),
    'Elements of list must be strings.'
)
listintervals = ListString(
    lambda x: (
        (isinstance(x, (list, tuple)))
        and (len(x) == 2)
        and (isinstance(x[0], Number))
        and (isinstance(x[1], Number))
    ),
    'Elements of list must be two-tuple of numbers (integers or float).'
)
dictstring = DictString()
dicttags = DictString(
    lambda x: isinstance(x, str),
    'Values of dict must be strings'
)
numberdict = DictString(
    lambda x: isinstance(x, Number),
    'Values of dict must be strings'
)
attachprocess = AttachProcess()
attachplaceholder = AttachPlaceholder()
lookupname = LookupName()
email = Email()
phone = Phone()
pickledata = PickleData()
pickleblob = PickleBlob()
folderpath = FolderPath()
stockid = PrefixId(prefix='s')

arrblob = ArrBlob()
arrstore = ArrStore()

arr1blob = ArrBlob(1)
arr1store = ArrStore(1)

arr2blob = ArrBlob(2)
arr2store = ArrStore(2)

arr3blob = ArrBlob(3)
arr3store = ArrStore(3)


custom_attributes_dict = {
    'chr': chr,
    'link': link,
    'flyidentifier': flyidentifier,
    'crossschema': crossschema,
    'truebool': truebool,
    'tarfolder': tarfolder,
    'liststring': liststring,
    'dictstring': dictstring,
    'dicttags': dicttags,
    'tags': tags,
    'attachprocess': attachprocess,
    'attachplaceholder': attachplaceholder,
    'lookupname': lookupname,
    'email': email,
    'phone': phone,
    'listintervals': listintervals,
    'pickledata': pickledata,
    'pickleblob': pickleblob,
    'folderpath': folderpath,
    'stockid': stockid,
    'dreyejson': dreyejson,
    'numberlist': numberlist,
    'numberdict': numberdict,
    'arrblob': arrblob,
    'arrstore': arrstore,
    'arr1blob': arr1blob,
    'arr1store': arr1store,
    'arr2blob': arr2blob,
    'arr2store': arr2store,
    'arr3blob': arr3blob,
    'arr3store': arr3store,
}



"""function for assigning fields
"""

import datetime
import os
import warnings
import pandas as pd
import numpy as np
import pickle
import json
import uuid

import datajoint as dj
from datajoint.declare import match_type
from datajoint import FreeTable
from datajoint.table import lookup_class_name
from datajoint.utils import to_camel_case
from flask import url_for
from wtforms import BooleanField, SelectField, DateField, DateTimeField, \
    StringField, FloatField, IntegerField, FormField, \
    TextAreaField, FieldList, DecimalField, HiddenField
from wtforms.validators import InputRequired, Optional, NumberRange, \
    ValidationError, Length, UUID, URL, Email
from werkzeug.datastructures import FileStorage
from markupsafe import escape

from loris import config
from loris.utils import is_manuallookup
from loris.app.utils import name_lookup, datareader
from loris.app.forms import NONES
from loris.app.forms.formmixin import (
    ManualLookupForm, ParentFormField, DynamicFileField, DictField, ListField,
    ParentValidator, JsonSerializableValidator, AttachFileField,
    BlobFileField, Extension, TagListField, MetaHiddenField,
    ParentInputRequired, Always, LookupNameValidator, PutValidator,
    HtmlLabelSelectWidget, TagDictField
)
from loris.database.attributes import PrefixId


class DynamicField:
    """Choose the right wtf field for each attribute in a datajoint Table

    Parameters
    ----------
    table : class
        datajoint.Table subclass representing a table in the mySQL database.
    attribute : `dj.Attribute`
        attribute in table to create a field for.
    ignore_foreign_fields : bool
        whether to process foreign keys to create foreign field forms or
        simply normal fields.
    """

    def __init__(self, table, attribute, ignore_foreign_fields=False):

        self._table = table
        self._attribute = attribute
        self._foreign_table, self._aliased = self.get_foreign_table()
        self._ignore_foreign_fields = ignore_foreign_fields

        # set default values
        self.foreign_fields = {}

    @property
    def is_integer(self):
        # won't work with adapted types
        return match_type(self.type) == 'INTEGER'

    @property
    def is_uuid(self):
        return match_type(self.type) == 'UUID'

    @property
    def is_datelike(self):
        return match_type(self.type) == 'TEMPORAL'

    @property
    def ignore_foreign_fields(self):
        return self._ignore_foreign_fields

    @property
    def aliased(self):
        return self._aliased

    @property
    def table(self):
        return self._table

    @property
    def attribute(self):
        return self._attribute

    @property
    def attr(self):
        if isinstance(self._attribute, list):
            return self._attribute[0]
        else:
            return self._attribute

    @property
    def name(self):
        if isinstance(self._attribute, list):
            return '___'.join([attr.name for attr in self._attribute])
        return self.attr.name

    @property
    def names(self):
        if isinstance(self._attribute, list):
            return [attr.name for attr in self._attribute]
        return [self.attr.name]

    @property
    def in_key(self):
        return self.attr.in_key

    @property
    def nullable(self):
        return self.attr.nullable

    @property
    def comment(self):
        if isinstance(self._attribute, list):
            return ''
        return self.attr.comment

    @property
    def default(self):
        if isinstance(self._attribute, list):
            return None
        return self.attr.default

    # shouldn't be used if attribute is foreign_key
    @property
    def type(self):
        return self.attr.type

    @property
    def sql_type(self):
        return self.attr.sql_type

    @property
    def adapter(self):
        return self.attr.adapter

    @property
    def is_prefixid(self):
        return (
            match_type(self.type) == 'ADAPTED'
        ) and isinstance(self.adapter, PrefixId)

    @property
    def is_blob(self):
        return self.attr.is_blob

    @property
    def dependencies(self):
        """
        """

        if not (self.table.connection.dependencies):
            self.table.connection.dependencies.load()

        return self.table.connection.dependencies

    @property
    def is_foreign_key(self):
        """whether attribute is foreign key
        """
        return self.foreign_table is not None

    def get_foreign_table(self):
        """
        """
        aliased = None
        parents = self.dependencies.parents(self.table.full_table_name)
        for table_name, table_info in parents.items():
            # TODO check
            if set(self.names) == set(table_info['attr_map']):
                # deal with aliasing
                if table_info['aliased']:
                    aliased_parents = self.dependencies.parents(table_name)
                    # aliased parent should only be one
                    table_name = list(aliased_parents)[0]
                    aliased = table_info['attr_map']
                break
        else:
            return None, aliased

        return config.get_table(table_name), aliased

    @property
    def foreign_table(self):
        return self._foreign_table

    @property
    def singular(self):
        return (
            not isinstance(self._attribute, list)
            or len(self._attribute) == 1
        )

    @property
    def foreign_data(self):
        if self.aliased is None:
            data = self.foreign_table.proj().fetch()
        else:
            data = self.foreign_table.proj(**self.aliased).fetch()

        if self.singular:
            data = data[self.name]
            # sort values if integer
            # if self.is_integer:
            # always sort values
            data = np.sort(data)
        else:
            data = pd.DataFrame(
                data
            ).sort_values(self.names)[self.names].to_records(index=False)
        # elif:
        #     data = pd.DataFrame(data).
        return data

    @property
    def foreign_comment_data(self):
        if self.foreign_is_simple_manuallookup:
            return self.foreign_table.proj("comments").fetch()["comments"]
        else:
            return self.foreign_data

    @property
    def foreign_is_manuallookup(self):
        if not self.is_foreign_key:
            return False
        return (
            (
                self.foreign_table.full_table_name
                != config.user_table.full_table_name
            ) & (
                self.foreign_table.full_table_name
                != config.assigned_table.full_table_name
            )
        )
        # return True # TODO test is_manuallookup(self.foreign_table)

    @property
    def foreign_is_simple_manuallookup(self):
        if not self.is_foreign_key:
            return False
        else:
            return is_manuallookup(self.foreign_table)

    def create_field(self):
        """create field for dynamic form
        """

        field = self._create_field()

        if field is None:
            warnings.warn(
                f'No field generated for {self.name} of '
                f'type {self.type}'
            )

        return field

    def _create_field(self, attr_type=None, kwargs=None):
        """create field for dynamic form
        """

        if kwargs is None:
            kwargs = self.get_init_kwargs()

        if (
            self.is_foreign_key
            and not self.ignore_foreign_fields
            and self.foreign_is_manuallookup
        ):
            return self.create_manuallookup_field(kwargs)
        elif (
            self.is_foreign_key
            # and (len(self.foreign_data) <= config['fk_dropdown_limit'])
        ):
            return self.create_dropdown_field(kwargs)

        if attr_type is None:
            type = match_type(self.type)
            sql_type = self.sql_type
        else:
            sql_type = attr_type
            type = match_type(attr_type)

        if type == 'INTEGER':
            return self.integer_field(kwargs)
        elif type == 'DECIMAL':
            return self.decimal_field(kwargs)
        elif type == 'FLOAT':
            return self.float_field(kwargs)
        elif type == 'STRING':
            return self.string_field(kwargs, sql_type)
        elif type == 'ENUM':
            return self.enum_field(kwargs, sql_type)
        elif type == 'BOOL':
            return self.bool_field(kwargs)
        elif type == 'TEMPORAL':
            return self.temporal_field(kwargs, sql_type)
        elif type in ('INTERNAL_BLOB', 'EXTERNAL_BLOB'):
            return self.blob_field(kwargs)
        elif type in ('INTERNAL_ATTACH', 'EXTERNAL_ATTACH'):
            return self.attach_field(kwargs)
        elif type == 'FILEPATH':
            return self.filepath_field(kwargs)
        elif type == 'UUID':
            return self.uuid_field(kwargs)
        elif type == 'ADAPTED':
            return self.adapted_field(kwargs)

    def get_init_kwargs(self):
        """get initialization dictionary to pass to field class
        """

        kwargs = {}
        attr_name = self.name.replace('___', ', ').replace('_', ' ')
        if not self.is_foreign_key:
            kwargs['label'] = attr_name
        else:
            # add table url
            url_kwargs = dict(zip(
                ['schema', 'table', 'subtable'],
                name_lookup(self.foreign_table.full_table_name).split('.')
            ))
            if self.in_key:
                color = 'Crimson'
            elif self.nullable:
                color = 'DarkGray'
            else:
                color = 'Black'
            kwargs['label'] = (
                f'<a href="{url_for("table", **url_kwargs)}" '
                f'target="_blank">'
                f'<font color={color}>'
                f'{attr_name}'
                '</font>'
                '</a>'
            )
        if self.comment.strip():
            kwargs['description'] = self.comment.strip()
        elif not self.is_foreign_key:
            kwargs['description'] = self.name.replace('___', ', ').replace('_', ' ')

        nullable = self.nullable  # or self.default in NONES
        kwargs['render_kw'] = {
            'nullable': self.nullable,
            'primary_key': self.in_key
        }

        # ignore_foreign_fields assumes that a parent form is being created
        if nullable or (self.ignore_foreign_fields and self.in_key):
            kwargs['default'] = None
        else:
            kwargs['default'] = self.default

        if (
            (self.in_key or not nullable)
            and not self.ignore_foreign_fields
        ):
            # case of no pop up
            kwargs['validators'] = [InputRequired()]
        elif (self.in_key or not nullable) and self.ignore_foreign_fields:
            # case when foreign key pop up form
            kwargs['validators'] = [ParentInputRequired()]
        else:
            kwargs['validators'] = [Optional()]

        return kwargs

    def integer_field(self, kwargs):
        # auto increment integer primary keys
        kwargs['default'] = self._get_integer_default()
        return IntegerField(**kwargs)

    def _get_integer_default(self):
        # auto increment integer primary keys
        default = None
        if self.ignore_foreign_fields and self.in_key:
            pass
        elif len(self.table()) == 0:
            default = 1
        elif self.in_key and len(self.table.heading.primary_key) == 1:
            default = np.max(
                self.table.proj().fetch()[self.name]
            ) + 1
        return default

    def _get_prefixid_default(self):
        default = None
        if self.ignore_foreign_fields and self.in_key:
            pass
        elif len(self.table()) == 0:
            default = f'{self.adapter.prefix}1'
        elif self.in_key and len(self.table.heading.primary_key) == 1:
            default = self.table.proj().fetch()[self.name]
            default = pd.Series(
                default
            ).str[len(self.adapter.prefix):].astype(int).max() + 1
            default = f'{self.adapter.prefix}{default}'
        return default

    def float_field(self, kwargs):
        return FloatField(**kwargs)

    def decimal_field(self, kwargs):
        # implement rounding correctly and use DecimalField
        return FloatField(**kwargs)

    def string_field(self, kwargs, sql_type):
        """creates a string field
        """
        max_length = int(sql_type.split('(')[-1][:-1])

        if sql_type.startswith('varchar'):
            kwargs['validators'].append(Length(max=max_length))
        elif sql_type.startswith('char'):
            kwargs['validators'].append(Length(min=max_length, max=max_length))
        else:
            return

        if max_length >= config['textarea_startlength']:
            return TextAreaField(**kwargs)
        else:
            return StringField(**kwargs)

    def temporal_field(self, kwargs, sql_type):
        """date and datetime field
        """

        if sql_type in ('datetime', 'timestamp'):
            if not (self.ignore_foreign_fields and self.in_key):
                kwargs['default'] = datetime.datetime.today
            kwargs['label'] += '&emsp;<small>(Y/m/d H:M - e.g. 2020/01/01 11:11)</small>&emsp;'
            return DateTimeField(format='%Y/%m/%d %H:%M', **kwargs)
        elif sql_type == 'time':
            if not (self.ignore_foreign_fields and self.in_key):
                kwargs['default'] = datetime.datetime.today
            kwargs['label'] += '&emsp;<small>(H:M - e.g. 11:11)</small>&emsp;'
            return DateTimeField(format='%H:%M', **kwargs)
        elif sql_type == 'date':
            if not (self.ignore_foreign_fields and self.in_key):
                kwargs['default'] = datetime.date.today
            kwargs['label'] += '&emsp;<small>(Y/m/d - e.g. 2020/01/01)</small>&emsp;'
            return DateField(format='%Y/%m/%d', **kwargs)
        elif sql_type == 'year':
            if not (self.ignore_foreign_fields and self.in_key):
                kwargs['default'] = datetime.date.today
            kwargs['label'] += '&emsp;<small>(Y - e.g. 2020)</small>&emsp;'
            return DateField(format='%Y', **kwargs)

    def _get_temporal_default(self):
        if self.sql_type in ('year', 'date'):
            if not (self.ignore_foreign_fields and self.in_key):
                return datetime.date.today
        elif not (self.ignore_foreign_fields and self.in_key):
            return datetime.datetime.today

    def enum_field(self, kwargs, sql_type):
        """create field for enum
        """
        choices = sql_type[sql_type.find('(')+1:sql_type.rfind(')')].split(',')
        choices = [ele.strip().strip('"').strip("'") for ele in choices]
        if self.nullable:
            choices = ['NULL'] + choices
        kwargs['choices'] = [(ele, ele) for ele in choices]
        return SelectField(**kwargs)

    def bool_field(self, kwargs):
        if kwargs['default'] is None:
            kwargs['default'] = False
        elif kwargs['default']:
            kwargs['default'] = True
        else:
            kwargs['default'] = False
        # input always optional
        kwargs['validators'][0] = Optional()
        return BooleanField(**kwargs)

    def blob_field(self, kwargs):
        kwargs['validators'] = [Optional()]  # TODO fix for editing
        kwargs['validators'].append(Extension())
        return BlobFileField(**kwargs)

    def attach_field(self, kwargs):
        kwargs['validators'] = [Optional()]  # TODO fix for editing
        kwargs['validators'].append(Extension(config['attach_extensions']))
        return AttachFileField(**kwargs)

    def filepath_field(self, kwargs):
        # TODO implement
        # kwargs['validators'].append(FilePath())
        return

    def uuid_field(self, kwargs):
        kwargs['validators'].append(Length(36, 36))
        kwargs['validators'].append(UUID())
        kwargs['default'] = str(uuid.uuid4())
        return StringField(**kwargs)

    def adapted_field(self, kwargs):
        """creates an adapted field type
        """

        try:
            attr_type = self.adapter.attribute_type
        except NotImplementedError:
            attr_type = self.sql_type

        attr_type_name = self.type.strip('<>')
        adapter = config['custom_attributes'].get(attr_type_name, None)

        if self.is_prefixid:
            kwargs['default'] = self._get_prefixid_default()

        if adapter is None:
            pass
        elif attr_type_name == 'liststring':
            kwargs['validators'].append(
                JsonSerializableValidator(list, adapter.put))
            return ListField(**kwargs)
        elif attr_type_name == 'dictstring':
            kwargs['validators'].append(
                JsonSerializableValidator(dict, adapter.put))
            return DictField(**kwargs)
        elif attr_type_name == 'tags':
            return TagListField(**kwargs)
        elif attr_type_name == 'dicttags':
            return TagDictField(**kwargs)
        elif attr_type_name == 'link':
            kwargs['validators'].append(URL(False))
        elif attr_type_name == 'email':
            kwargs['validators'].append(Email())
        elif attr_type_name == 'lookupname':
            kwargs['validators'].append(LookupNameValidator())
        else:
            kwargs['validators'].append(PutValidator(adapter.put))

        return self._create_field(attr_type, kwargs)

    def create_manuallookup_field(self, kwargs):
        """create a manual lookup field form.
        """

        kwargs['id'] = 'existing_entries'
        if self.aliased is None:
            kwargs['validators'].insert(0, ParentValidator(self.names))
        else:
            kwargs['validators'].insert(
                0,
                ParentValidator([self.aliased[name] for name in self.names])
            )

        # dynamically create form
        class FkForm(ManualLookupForm):
            parent_table_name = to_camel_case(self.foreign_table.table_name)
            existing_entries = self.create_dropdown_field(kwargs)

        for name, attr in self.foreign_table.heading.attributes.items():
            dynamicfield = DynamicField(self.foreign_table, attr, True)
            # create field
            field = dynamicfield.create_field()
            if field is not None:
                setattr(
                    FkForm,
                    name,
                    field
                )
                # keep track of dynamic fields
                self.foreign_fields[name] = dynamicfield

        return ParentFormField(FkForm)

    def create_dropdown_field(self, kwargs):
        """a simple drowndown field for foreign keys
        """

        choices = self.get_foreign_choices()
        if self.nullable:
            kwargs['default'] = 'NULL'

        if choices:
            kwargs['choices'] = choices
            kwargs['widget'] = HtmlLabelSelectWidget()
            return SelectField(**kwargs)
        else:
            kwargs['label'] += ' -- <font color="red">MISSING ENTRIES IN PARENT TABLE!</font>'
            kwargs['validators'].append(Always())
            return MetaHiddenField(**kwargs)

    def get_foreign_choices(self):
        if self.foreign_is_simple_manuallookup:
            choices = [
                (str(ele), (str(ele), str(comment)))
                if comment is not None
                else
                (str(ele), str(ele))
                for ele, comment
                in zip(self.foreign_data, self.foreign_comment_data)
            ]
        else:
            choices = [
                (str(ele), str(ele))
                for ele in self.foreign_data
            ]
        choices = sorted(choices)
        if self.nullable:
            choices = [('NULL', 'NULL')] + choices
        if self.foreign_is_manuallookup and not self.ignore_foreign_fields:
            choices += [('<new>', '<add new entry>')]

        return choices

    def format_value(self, value):
        """format value
        """

        if self.foreign_is_manuallookup:
            if value['existing_entries'] == '<new>':
                value.pop('existing_entries')
                try:
                    self.foreign_table.insert1(
                        self.foreign_table_format_value(value)
                    )
                except dj.DataJointError as e:
                    raise dj.DataJointError(
                        "An error occured while inserting into parent table"
                        f" {self.foreign_table.full_table_name}: {e}"
                    )
                if self.aliased is None:
                    return {name: value[name] for name in self.names}
                else:
                    return {name: value[self.aliased[name]]
                            for name in self.names}
            elif self.singular:
                return {
                    self.name: value['existing_entries']
                }
            else:
                values = value['existing_entries'].strip('()').split(', ')
                # strip any quotes
                values = [ele.strip().strip('"').strip("'") for ele in values]
                return dict(zip(self.names, values))

        if self.is_blob:
            value = self._process_blob_value(value)

        # return value
        return {self.name: value}

    @staticmethod
    def _process_blob_value(value):
        """process blobs by loading files
        """

        if value in NONES:
            value = None
        else:
            value = datareader(value)

        return value

    def foreign_table_format_value(self, value):
        """process foreign table values before insertion
        """
        for name, attr in self.foreign_table.heading.attributes.items():
            if attr.is_blob and name in value:
                value[name] = self._process_blob_value(value[name])
        return value

    def prepare_populate(self, value):
        """format value for populating form
        """

        if self.foreign_is_manuallookup:
            value = {
                'existing_entries': str(value)
            }

        if self.is_blob and value is not None:
            # create filepath
            # TODO - not functional atm
            filepath = os.path.join(
                config['tmp_folder'],
                str(uuid.uuid4()) + '.pkl'
            )
            with open(filepath, 'wb') as f:
                pickle.dump(value, f, protocol=pickle.HIGHEST_PROTOCOL)

            return filepath

        return value

    def update_field(self, form):

        if self.foreign_is_manuallookup and not self.ignore_foreign_fields:
            formfield = getattr(form, self.name)
            formfield.existing_entries.choices = self.get_foreign_choices()

            # update fields in parent form
            for name, foreign_dynamicfield in self.foreign_fields.items():
                foreign_dynamicfield.update_field(formfield)

            # update field if necessary
            if self.singular:
                self._update_field(formfield)

        elif self.is_foreign_key:
            field = getattr(form, self.name)
            field.choices = self.get_foreign_choices()

        else:
            self._update_field(form)

    def _update_field(self, form):

        if self.is_uuid:
            field = getattr(form, self.name)
            field.default = str(uuid.uuid4())
        if self.is_integer:
            field = getattr(form, self.name)
            field.default = self._get_integer_default()
        if self.is_datelike:
            field = getattr(form, self.name)
            field.default = self._get_temporal_default()
        if self.is_prefixid:
            field = getattr(form, self.name)
            field.default = self._get_prefixid_default()
